import static java.lang.Math.random;

/**
 * This class implements a max heap of words using an array.  The methods to complete heapsort are
 * included.
 *
 * @author Michael Dube
 * @version 1.0
 * @since 2019-11-01
 */
class MaxHeapWords {

    private Word[] values; // Words stored in the heap
    private static final int MEAN_WORD_SIZE = 3; //

    /**
     * Creates a max heap of num_words words.  Thus the words will be in alphabetical order after
     * being sorted.
     *
     * @param num_words Number of words to generate
     */
    MaxHeapWords(int num_words) {

        values = new Word[num_words];
        rndValues(num_words);
        print(values, "Before Sorting: ");
        sort(values);
        print(values, "After Sorting: ");

    } // constructor

    /**
     * Fills the values array with num_words randomly generated words.
     *
     * @param num_words Number of words to generate
     */
    private void rndValues(int num_words) {
        for (int idx = 0; idx < num_words; idx++) {
            values[idx] = new Word((int) (random() * 4 + MEAN_WORD_SIZE));
        }
    } // rndValues

    /**
     * This method takes in an unsorted array of words and sorts the array in alphabetical order
     * using heap sort.
     *
     * @param arr Array to be sorted
     */
    private void sort(Word[] arr) {
        int numItems = arr.length;
        Word swap;                   // For swapping values within array

        // First, build the heap
        for (int i = numItems / 2 - 1; i >= 0; i--) {
            heapify(arr, numItems, i);
        }

        // Move max (root) to end of array, for each value within the array
        for (int i = numItems - 1; i >= 0; i--) {
            swap = arr[0];
            arr[0] = arr[i];
            arr[i] = swap;

            // Ensure the remainder of the heap satisfies Max Heap Property
            heapify(arr, i, 0); // Only considers arr[0:i]
        }
    } // sort

    /**
     * This method is used to ensure that a subtree rooted with node i, an index within arr[],
     * satisfies the Max Heap Property.  This property dictates that the parent node must be
     * 'higher' or 'equal' in alphabetical order (later in the alphabet) in comparison to both
     * children.
     *
     * @param arr Array holding heap
     * @param n   Size of the heap
     * @param i   Root index of subtree being considered
     */
    private void heapify(Word[] arr, int n, int i) {
        int maxIndex = i;               // Index of Max Value, initialized as root
        int leftChild = 2 * i + 1;      // Left Child index of Root
        int rightChild = 2 * i + 2;     // Right Child index of Root
        Word swap;                      // For swapping values within array

        // Left Child Larger than Root
        if (leftChild < n && arr[leftChild].wordVal() > arr[maxIndex].wordVal()) {
            maxIndex = leftChild;
        }

        // Right Child Larger than root
        if (rightChild < n && arr[rightChild].wordVal() > arr[maxIndex].wordVal()) {
            maxIndex = rightChild;
        }

        if (maxIndex != i) { // Largest value isn't root (violating max heap property)
            swap = arr[i];
            arr[i] = arr[maxIndex];
            arr[maxIndex] = swap;

            // Check sub-tree
            heapify(arr, n, maxIndex);
        }
    } // heapify

    /**
     * This method prints an array to console with specified beginning message.
     *
     * @param arr Array to print
     * @param msg Message before printed array
     */
    private void print(Word[] arr, String msg) {
        System.out.print(msg);
        for (Word x : arr) {
            System.out.print(x.spelling + "\t");
        }
        System.out.println();
    } // print

} // MaxHeapGeneric
